import { LitElement } from 'lit-element';
/** A rule for configuring how to sort by an attribute or property value */
export interface Rule {
    /** Attribute name, or array representing a path of properties. (e.g. el.innerText -> ['innerText'] or el.someObj.someChild.someGrandchild -> ['someObj', 'someChild', 'someGrandchild'])  </br>
     * Note: Changes to values of sorted attributes will trigger a re-sort. Changes to sorted properties will not.
    */
    key: string | string[];
    /** Selector for descendant to get attribute/property off of */
    selector?: string;
    /** If true, sort in reverse order relative to the global sort direction */
    reverse?: boolean;
}
/**
 * A web component for sorting elements
 *
 * @element ds-sorter
 *
 * @slot - Content to sort
 */
export declare class DsSorter extends LitElement {
    #private;
    /**
     * If present, sorts randomly
     */
    random: boolean;
    /**
     * A list of comma-separated rules to sort by in order of precedence. <br/>Specify attributes by name (e.g. "href"). If specifying a property, prepend with "." (e.g. ".innerText"). You can use nested properties as well (e.g. ".dataset.row"). <br/>Optionally, if you'd like to reverse a rule relative to the others, prepend a ">" (e.g. "href, >.innerText"). <br/>Finally, if you'd like to get a value of a descendant of the sorted element, wrap a selector in braces before the value and modifiers (e.g. {div label input} .checked).
     */
    get by(): string;
    set by(newBy: string);
    /**
     * A list of rule objects to sort the elements by. Refer to Rule interface for properties.
     */
    rules: Rule[];
    /**
     * Custom [comparison function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/sort) for sorting
     */
    comparator: undefined | ((a: HTMLElement, b: HTMLElement) => number);
    /**
     * Sort in descending order (else ascending is default)
     */
    descending: boolean;
    private get _slottedContent();
    disconnectedCallback(): void;
    updated(): void;
    render(): import("lit-element").TemplateResult;
    /**
     * @method
     * Manually trigger a sort, such as in response to an event e.g. <ds-sorter onchange="this.sort()">...</ds-sort>
     */
    sort(): void;
}
