var _mutationObserver, _elementAttrsMap, _compareElements, _getValue;
import { __classPrivateFieldGet, __decorate } from "tslib";
import { html, LitElement, property } from 'lit-element';
const parseToRules = (value) => {
    if (!value)
        return [];
    // TODO: Validate with regex?
    const stringRules = value.split(/,\s*(?![^{}]*\})/);
    return stringRules.map(stringRule => {
        const [rawKey, selector] = stringRule.replace('{', '').split(/\}\s*/).reverse();
        let key = rawKey;
        const reverse = key[0] === '>';
        if (reverse)
            key = key.slice(1);
        if (key[0] === '.') {
            [, ...key] = key.split('.');
        }
        // Default to .innerText if selector and/or reverse but no key 
        if (key === undefined || key === '') {
            key = ['innerText'];
        }
        return {
            key,
            selector,
            reverse,
        };
    });
};
/**
 * A web component for sorting elements
 *
 * @element ds-sorter
 *
 * @slot - Content to sort
 */
export class DsSorter extends LitElement {
    constructor() {
        super(...arguments);
        /**
         * If present, sorts randomly
         */
        this.random = false;
        /**
         * A list of rule objects to sort the elements by. Refer to Rule interface for properties.
         */
        this.rules = [{ key: ['innerText'] }];
        /**
         * Custom [comparison function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/sort) for sorting
         */
        this.comparator = undefined;
        /**
         * Sort in descending order (else ascending is default)
         */
        this.descending = false;
        _mutationObserver.set(this, new MutationObserver(mutationList => {
            const shouldUpdate = mutationList.some(mutation => mutation.type === 'attributes');
            if (shouldUpdate)
                this.sort();
        }));
        _elementAttrsMap.set(this, new WeakMap());
        _compareElements.set(this, (a, b, rules = this.rules) => {
            if (this.random) {
                return Math.random() - 0.5;
            }
            if (this.comparator) {
                return this.comparator(a, b) * (this.descending ? -1 : 1);
            }
            const [rule, ...restRules] = rules;
            // No rule found, don't sort
            if (!rule) {
                return 0;
            }
            const { reverse = false } = rule;
            const firstVal = __classPrivateFieldGet(this, _getValue).call(this, a, rule);
            const secondVal = __classPrivateFieldGet(this, _getValue).call(this, b, rule);
            const lesser = (this.descending != reverse) ? 1 : -1;
            const greater = -lesser;
            if ((firstVal == undefined && secondVal == undefined) || firstVal === secondVal) {
                // If current values are equal, move down the rules until something isn't equal or we run out of rules
                return restRules.length && __classPrivateFieldGet(this, _compareElements).call(this, a, b, restRules);
            }
            // send nullish values to the end
            if (firstVal == undefined && secondVal != undefined) {
                return greater;
            }
            if (firstVal != undefined && secondVal == undefined) {
                return lesser;
            }
            if (firstVal < secondVal) {
                return lesser;
            }
            return greater;
        }
        /** Normalize value for comparison */
        );
        /** Normalize value for comparison */
        _getValue.set(this, (sortingElem, rule) => {
            var _a, _b, _c;
            const { key, selector } = rule;
            const elem = (selector ? sortingElem.querySelector(selector) : sortingElem);
            if (elem === null) {
                console.warn(`ds-sorter: Selector ${selector} did not return an element`);
                return undefined;
            }
            if (typeof key === 'string')
                return elem.getAttribute(key);
            const [firstProp, ...nestedProps] = key;
            if (!(firstProp in elem)) {
                console.warn(`ds-sorter: Element does not have property '${firstProp}'`);
                return undefined;
            }
            let prop = elem[firstProp];
            let prevProp = firstProp;
            for (const nestedProp of nestedProps) {
                if (prop == undefined || typeof prop !== 'object') {
                    console.warn(`ds-sorter: Cannot access nested property '${nestedProp}' on element because property '${prevProp}' is not an object`, elem);
                    return undefined;
                }
                if (!(nestedProp in prop)) {
                    console.warn(`ds-sorter: Element property '${prevProp}' does not contain nested property '${nestedProp}'`, elem);
                    return undefined;
                }
                prop = prop[nestedProp];
                prevProp = nestedProp;
            }
            // For now, just treat NaN as undefined since it doesn't compare nicely with anything, including itself
            if (typeof prop === 'number' && isNaN(prop)) {
                return undefined;
            }
            const returnAsIs = ['number', 'string', 'boolean', 'bigint', 'undefined'];
            if (returnAsIs.includes(typeof prop) || prop === null) {
                return prop;
            }
            // TS doesn't think this could ever be a symbol (e.g. someone sets a symbol as the value of a property on an element. Why would anyone do this? I don't know, but they can if they want to.)
            if (typeof prop === 'symbol') {
                const { description } = prop;
                console.warn(`The value being sorted by is a symbol. Using symbol description: "${description}".`);
                return description;
            }
            if (typeof prop === 'function') {
                console.warn('The value being sorted by is a function. Using value "true".');
                // No good way to sort functions, just return that it exists
                return true;
            }
            if (typeof prop === 'object') {
                // If array-like, return the length, else get the valueOf value if it's implemented
                return (_b = (_a = prop) === null || _a === void 0 ? void 0 : _a.length) !== null && _b !== void 0 ? _b : (_c = prop === null || prop === void 0 ? void 0 : prop.valueOf) === null || _c === void 0 ? void 0 : _c.call(prop);
            }
            return undefined;
        });
    }
    /**
     * A list of comma-separated rules to sort by in order of precedence. <br/>Specify attributes by name (e.g. "href"). If specifying a property, prepend with "." (e.g. ".innerText"). You can use nested properties as well (e.g. ".dataset.row"). <br/>Optionally, if you'd like to reverse a rule relative to the others, prepend a ">" (e.g. "href, >.innerText"). <br/>Finally, if you'd like to get a value of a descendant of the sorted element, wrap a selector in braces before the value and modifiers (e.g. {div label input} .checked).
     */
    get by() {
        return this.rules.map(rule => `${rule.selector ? `{${rule.selector}}` : ''} ${rule.reverse ? '>' : ''}${typeof rule.key === 'string' ? rule.key : '.' + rule.key.join('.')}`).join(', ');
    }
    set by(newBy) {
        // Will trigger update via rules prop, no need to do it here
        this.rules = parseToRules(newBy);
    }
    get _slottedContent() {
        var _a, _b, _c;
        return Array.from((_c = (_b = (_a = this.shadowRoot) === null || _a === void 0 ? void 0 : _a.querySelector('slot')) === null || _b === void 0 ? void 0 : _b.assignedElements()) !== null && _c !== void 0 ? _c : []);
    }
    disconnectedCallback() {
        super.disconnectedCallback();
        __classPrivateFieldGet(this, _mutationObserver).disconnect();
    }
    updated() {
        this.sort();
        this._slottedContent.forEach(elem => {
            this.rules.forEach(rule => {
                var _a;
                const { key, selector } = rule;
                if (typeof key === 'string') {
                    const observeElem = (selector ? elem.querySelector(selector) : elem);
                    // Will throw console warning in .sort() if selector doesn't return element
                    if (observeElem) {
                        const attributeSet = (_a = __classPrivateFieldGet(this, _elementAttrsMap).get(observeElem)) !== null && _a !== void 0 ? _a : new Set();
                        attributeSet.add(key);
                        __classPrivateFieldGet(this, _elementAttrsMap).set(observeElem, attributeSet);
                        __classPrivateFieldGet(this, _mutationObserver).observe(observeElem, { attributeFilter: Array.from(attributeSet) });
                    }
                }
            });
        });
    }
    render() {
        return html `
      <slot @slotchange=${this.sort}></slot>
    `;
    }
    /**
     * @method
     * Manually trigger a sort, such as in response to an event e.g. <ds-sorter onchange="this.sort()">...</ds-sort>
     */
    sort() {
        this._slottedContent.sort(__classPrivateFieldGet(this, _compareElements))
            .forEach((el, i) => { var _a, _b; return (((_a = el.parentElement) === null || _a === void 0 ? void 0 : _a.children[i]) !== el) && ((_b = el.parentElement) === null || _b === void 0 ? void 0 : _b.appendChild(el)); });
        this.dispatchEvent(new CustomEvent('ds-sort', { composed: true, bubbles: true }));
    }
}
_mutationObserver = new WeakMap(), _elementAttrsMap = new WeakMap(), _compareElements = new WeakMap(), _getValue = new WeakMap();
__decorate([
    property({ type: Boolean })
], DsSorter.prototype, "random", void 0);
__decorate([
    property()
], DsSorter.prototype, "by", null);
__decorate([
    property({ type: Array })
], DsSorter.prototype, "rules", void 0);
__decorate([
    property({ attribute: false })
], DsSorter.prototype, "comparator", void 0);
__decorate([
    property({ type: Boolean })
], DsSorter.prototype, "descending", void 0);
//# sourceMappingURL=DsSorter.js.map